#include "ConstraintDB.hpp"
#include "Settings.hpp"
#include <algorithm>


void ConstraintDB::clsCreateBinary(const Literal lit0, const Literal lit1){

asrt(lit0.idx() < watches.size() && lit1.idx() < watches.size());

    watches[lit0.idx()]   .push_back(WatchedElement(lit1));
   	watches[lit1.idx()]   .push_back(WatchedElement(lit0));
   	++nbr_bin_clses;
}



Clause* ConstraintDB::clsCreate(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend, bool learnt,bool link){

   // int32_t id(clauses.size());
    uint32_t cl_sz							= lend-lbeg;
    asrt(cl_sz > 2);

    void* mem								= malloc(sizeof(Clause) + sizeof(Literal)*cl_sz);
    if(mem == NULL){
    	fprintf(stderr,"c  !!!OUT OF MEMORY - Try to Fix!!!\n");
    	exit(EXIT_FAILURE);
    	return NULL;
    }
    Clause* cls 							= new (mem) Clause(cl_sz, learnt);

    Literal lit0 							= *lbeg;
    Literal lit1 							= *(lbeg+1);
    for(uint32_t i = 0;lbeg != lend;++lbeg) {(*cls)[i++] = *lbeg;}

    if(learnt)  {
    	learnt_cls                          .push_back(cls);
        nbr_long_learnt_literals			+= cl_sz;
    }else{
        fix_cls                             .push_back(cls);
        nbr_long_fix_literals				+= cl_sz;
    }


    if(link){
    	asrt(lit0.flipped().idx() < watches.size() && lit1.flipped().idx() < watches.size());
    	clsLinkWatched(cls);
    }
    asrt(learnt == cls->isLearnt() && cls->isWatched() == link);

    return cls;
}



void    ConstraintDB::clsUncreate(Clause* cl, bool unlink, bool free_mem){

	asrt(!cl->isRubbish());
    asrt(cl->size() >= 2);
    asrt(unlink == cl->isWatched());
    asrt(unlink || free_mem);


    if(cl->isLearnt())	nbr_long_learnt_literals 	-= cl->size();
    else				nbr_long_fix_literals 		-= cl->size();

    if(unlink){
    	WatchedElement::ClauseEqualFunctor cl_equal(cl);
    	asrt(find(litWatchedClauses((*cl)[0]),cl_equal));
    	asrt(find(litWatchedClauses((*cl)[1]),cl_equal));

    	clsUnlinkWatched(cl,false);
    }

    if(free_mem){
    	free(cl);
    }else{
    	cl->setRubbish();
    }
}



void ConstraintDB::reserveForVariables(uint32_t need_sz){

    if(need_sz > assignments.size()){
        //Variable infos
        assignments     .resize(need_sz,UNKNOWN_TrBool.toInt());
        assign_info     .resize(need_sz,AssignInfo());

        //Literal infos
        need_sz = need_sz << 1;
        watches         .resize(need_sz);
    }
}


void ConstraintDB::learntsGC(uint32_t new_sz,const bool free_mem){

	//print_out("new size %d, old size %d \n",new_sz,learnt_cls.size());
	if(new_sz >= learnt_cls.size()){
		//print_out("Would like to shrink but size is %d and you want %d \n",learnt_cls.size(),new_sz);
		return;
	}

    statistics.nbr_learnt_reductions += 1;

    //const uint32_t nbr_bins  = numberOfLearntBinClauses();

    std::nth_element(learnt_cls.begin(), learnt_cls.begin() + new_sz -1,
                       learnt_cls.end(), Clause::ActivityDescendingFunctor());

    //print_out("Deletion start at %u, number bin cls is %u and number of current learnts is %u \n",new_sz,nbr_bins,learnt_cls.size());
    for(uint32_t del_idx = new_sz; del_idx < learnt_cls.size();++del_idx){
    	Clause* cls = learnt_cls[del_idx];
    	if(clsLocked(*cls) || !cls->isWatched()){
        //if(clsLocked(*cls) || cls->filterFlagged() || !cls->isWatched()){
        	learnt_cls[new_sz++] = cls;
        }else{
        	clsUncreate(learnt_cls[del_idx],true,free_mem);
        }
    }


    learnt_cls.resize(new_sz);
}

/*
struct ProblemRelevantDescendingFunctor {
	bool operator() (Clause* a, Clause* b){
		return 	a->filterFlagged() == b->filterFlagged()?
				a->getActivity() > b->getActivity() :
            	a->filterFlagged();}
};*/

/*
uint32_t ConstraintDB::keepProblemClauses(uint32_t cut, std::list<Clause*>& out_container){

	asrt(false);
	return 0;

	if(cut >= numberOrigLongClses()) return cut = numberOrigLongClses();
	asrt(out_container.empty());

	cls_iter_t start = fix_cls.begin() + cut;

    std::nth_element(fix_cls.begin(), start, fix_cls.end(), ProblemRelevantDescendingFunctor());

    asrt(!(*start)->filterFlagged() || debugRangeFilterFlagged(fix_cls.begin(),start));

    for(;start != fix_cls.end();++start){
    	Clause* cl = (*start);

    	if(cl->filterFlagged()){++cut; continue;}

    	out_container.push_back(cl);
        WatchedElement::ClauseEqualFunctor cl_equal(cl);
        asrt(find(litWatchedClauses((*cl)[0]),cl_equal));
        asrt(find(litWatchedClauses((*cl)[1]),cl_equal));

    	//clsUnlinkWatched(cl);
    }
    return cut ;
}

*/

bool ConstraintDB::clsCanBeOmitted(Clause& cl){


	for(uint32_t i = 0; i != cl.size(); ) {
		if(litIsTrue(cl[i])){ 							return true;}
		if(litIsFalse(cl[i]) && i > 1){ // if i <= 1 -> clause is unit and other watcher should be true
			clsRemoveLiteral(cl,i);
			continue;
		}
		++i;
	}

	const uint32_t sz = cl.size(); asrt(sz > 1);

	if(sz > 2) 											return false;

	clsCreateBinary(cl[0],cl[1]);						return true;
}




void ConstraintDB::removeSat(vector<Clause*>& cls){
    vector<Clause*>::iterator to, iter;
    for (iter = to = cls.begin(); iter != cls.end(); ++iter){
        if ((*iter)->isWatched() && clsCanBeOmitted(**iter)){
        	statistics.nbr_simplify_worked += 1;
            clsUncreate(*iter,true,true);
        }else{
        	asrt(!(*iter)->isWatched() || (litIsUnknown((**iter)[0])&& litIsUnknown((**iter)[1])));
            *to++ = *iter;
        }
    }
    cls.erase(to,cls.end());
}


void ConstraintDB::printModel(){
	printf("v ");
	const uint32_t sz = numberVariables();
	for(uint i = 0; i != sz;++i){
		Variable v(i);
		TrBool val = varValue(v);
		//if(val.isUnknown()) continue;
		printf("%d ",(val.isFalse()?v.toNegLit().toDIMACS():v.toPosLit().toDIMACS()));
	}
	printf(" 0\n");
}


void ConstraintDB::printHookModel(){
	printf("v ");
	const uint32_t sz = numberVariables();
	for(uint i = 0; i != sz;++i){
		Variable v(i);
		TrBool val = varHook(v);
		//if(val.isUnknown()) continue;
		printf("%d ",(val.isFalse()?v.toNegLit().toDIMACS():v.toPosLit().toDIMACS()));
	}
	printf(" 0\n");
}






bool ConstraintDB::debugVerifyModel(){

    for(int i(fix_cls.size());i--;){
        if(checkSatByLit(*(fix_cls[i])) != INVALID_LIT) continue;

        if(settings.console_out != NULL){
        	settings.console_out->writeString("|Found unsatisfied clause: ");
            Clause& cl = *(fix_cls[i]);
            for(uint32_t j = 0; j != cl.size(); ++j) {
                Literal lit = cl[j];
                settings.console_out->writeString(" %i: %s",lit.toDIMACS(),(litIsFalse(lit)?"false":"unknown"));
            }
        }

        return  false;
    }

    for(uint32_t w=watches.size();w--;){
    	Literal lit(w); if(litIsTrue(lit)) continue;
    	we_vec& watched    = litWatchedClauses(lit);

    	for(we_iter iter = watched.begin();iter !=watched.end();++iter){
   	    	 WatchedElement& we = *iter;
   	    	 if(we.isBinary()){
   	    		 if(!litIsTrue(we.getLiteral())){
   	    			 std::cerr << " binary clause not sat " << lit.toDIMACS() << " " << we.getLiteral().toDIMACS() << std::endl;
   	    			 return false;
   	    		 }
   	    	 }
    	}

    }

    if(settings.console_out != NULL) settings.console_out->writeString("|Verified %d original clauses ",fix_cls.size());
    return true;
}


bool ConstraintDB::debugVerifyModelByHooks(){


    for(int i(numberOrigLongClses());i--;){
        if(checkSatByHook(*(fix_cls[i])) != INVALID_LIT) continue;

        if(settings.console_out != NULL){
        	settings.console_out->writeString("|Found unhooked clause: ");
            Clause& cl = *(fix_cls[i]);
            for(uint32_t j = 0; j != cl.size(); ++j) {
                Literal lit = cl[j];
                settings.console_out->writeString(" %i: %s",lit.toDIMACS(),(litIsHooked(lit)?"true":"false"));
            }
            settings.console_out->writeString(" Clause is%s learnt \n",((cl.isLearnt()?"":" not")));

            WatchedElement::ClauseEqualFunctor cl_equal(&cl);
            bool first 	= find(litWatchedClauses(cl[0].flipped()),cl_equal);
            bool second = find(litWatchedClauses(cl[1].flipped()),cl_equal);
            settings.console_out->writeString(" Can%s find first watched and can%s find second watched\n",(first?"": "not"),((second?"":" not")));
        }


        return false;
    }

    if(settings.console_out != NULL) settings.console_out->writeString("|Verified %d hooked clauses ",numberOrigLongClses());
    return true;
}


/*
bool ConstraintDB::debugCheckAssignReason(){
	EXPENSIVE_DEBUG

    for(uint32_t i(assign_info.size());i--;){
        if(assign_info[i].noReason() || assign_info[i].isBinReason()) continue;

        asrt(assign_info[i].getClsReason() != NULL);
        Clause& c = *(assign_info[i].getClsReason());
        Variable v(i);
        if(c[0].toVar() != v && c[1].toVar() != v){
        	if(settings.console_out != NULL){
        		settings.console_out->writeString("Variable %u has invalid assign reason!",v.toDIMACS());}
        	RETURN(false);
        }
    }

    RETURN(true);
}*/


const char*  constraintdb_cpp_version(){return "ConstraintDB.cpp:   $Revision: 1.3 $";}


