#include "DMRPSolver.hpp"
#include <algorithm>

struct ClActAscendingFunctor {
	bool operator() (Clause* a, Clause* b){
		asrt(!a->isRubbish() && !b->isRubbish());
	    return 	a->isLearnt() == b->isLearnt()?
	    	a->getActivity() < b->getActivity() :
	        b->isLearnt();}
};


void	DMRPSolver::printModel(){
	asrt(state_ok);

	if(settings.simp)
	{
		vector<TrBool> solverAssignments;
		solverAssignments.resize(0);
		for(uint i=0;i<numberVariables();i++){
			/*if(last_solver_type == DMRP)
				solverAssignments.push_back(varHook(Variable(i)));
			else*/
				solverAssignments.push_back(varValue(Variable(i)));
//				if(i>3710 && i<3740){
//					TrBool val = varValue(Variable(i));
//					cerr << "value of " << i << " is " << (val.isKnown()?(val.isTrue()?"t":"f"):"u") << endl;
//				}

		}

		(*sapplifyalot).compAssignments(solverAssignments);
	}
	else
	{
		if(last_solver_type == DMRP){
			printf("c DMRP Model\n");
			//print_out("Hier muss das model raus\n");
			ConstraintDB::printHookModel();
		}else {
//			asrt(assign_stack.size() == numberVariables());
			printf("c Basic Model\n");
			//print_out("Hier muss das model raus\n");
			ConstraintDB::printModel();
		}
	}
}


void DMRPSolver::init(){
	var_flags		.ensureSize(numberVariables(),false);
	hook_responses	.resize(numberVariables());
	var_makes		.resize(numberVariables());
	dec_var_count	.resize(numberVariables(),0);
	conflict_count	.setNumberClauses((double)numberAllClauses());
	luby_restarts	.setScale(100);
}

TrBool 	DMRPSolver::isSAT( uint32_t timeout){

	if (!state_ok)          return FALSE_TrBool;
	init();

	 if(settings.console_out != NULL){
	 		settings.console_out->writeCenteredToBG("Hybrid DMRP SApperloT ... YOU GOT ME SEARCHING ... ");
	 		settings.console_out->writeCenteredToBG(basename(settings.cnf_file_name));

			//TODO: simplifier:
			settings.console_out->writeCenteredToBG("simplifying started ...");
	 }

	 if(settings.simp){

		 statistics.nbr_clauses_original = currentId;
		 //statistics.nbr_literals_original will be set in the parsing method...

		 statistics.nbr_clauses_simplified = (*sapplifyalot).getActiveClausecount();
		 statistics.nbr_literals_simplified = (*sapplifyalot).getLiteralCount();

		 //simplifying with threshold
		 if((*sapplifyalot).getActiveClausecount() < 3000000){
			 (*sapplifyalot).simplify(settings.simptimeout);
			 if (!state_ok)	return FALSE_TrBool;

			 statistics.nbr_block_clauses = (*sapplifyalot).getAmountBlockingClauses();
			 statistics.nbr_clauses_simplified = (*sapplifyalot).getActiveClausecount();
			 statistics.nbr_literals_simplified = (*sapplifyalot).getLiteralCount();
		 }

		 //getting the simplified constraints
		 (*sapplifyalot).exportConstraintsToSolver();

		 statistics.simplifying_timestmp = statistics.cpuUserTime();
		 statistics.nbr_block_clauses = (*sapplifyalot).getAmountBlockingClauses();

		 //set number of clauses for DB reduction
		 conflict_count.setNumberClauses( statistics.nbr_clauses_simplified ); //+ (*sapplifyalot).getVarAssignmentsCount() );

	 }else{
		 statistics.nbr_clauses_simplified = numberAllClauses();
		 statistics.nbr_literals_simplified = numberLiterals();
	 }

	 if(settings.console_out != NULL){
		 settings.console_out->writeCenteredToBG("SAT Solver started ...");
			statistics.printStatsProgressHead();
			settings.console_out->writeString(
				"|____________________________________________________________________________________________|\n");
	 }

	if (!state_ok)		return FALSE_TrBool;
	TrBool status 		= UNKNOWN_TrBool;
	while(statistics.cpuUserTime() <= timeout){

		uint64_t rel_max_confl 	= luby_restarts.nextRestartGetConflictCount();
		bool dmrp 				= settings.hybrid && luby_restarts.isMinimumConflicts(rel_max_confl);


		if(dmrp){
			if(last_solver_type == BASIC){
				if(!computeHooksEnsureBinariesSAT()) return TrBool(state_ok = false);
			}else{
				backjumpKeepLevel(0,BACKJUMP_SET_PREV);
			}

			//print_out("Enter dmrp: %d\n",unwatched_clauses.size());
			status = dmrpSolveConsiderTimeAndConflicts(timeout,rel_max_confl );
			//print_out("Leave dmrp: %d\n",unwatched_clauses.size());
			statistics.nbr_dmrp_calls++;

			if(status.isTrue()){
				backjumpKeepLevel(0,BACKJUMP_SET_PREV);
				for(uint32_t i=numberVariables();i--;){
					Variable v(i);
					TrBool vval	= varValue(v);
					if(vval.isUnknown()){
						TrBool hval = varHook(v);
						if(hval.isKnown()){uncheckedEnqueue<Clause*>(hval.isTrue()?v.toPosLit():v.toNegLit(),NULL);}
					}
				}
			}

			last_solver_type = DMRP;
		}else{

			backjumpKeepLevel(0,BACKJUMP_SET_PREV);
			if(!relinkAllUnwatchedClauses()){ return TrBool(state_ok = false);}
			//resetNormalClsState();

			status = search( INIT, conflict_count.getConflicts() + rel_max_confl);
			last_solver_type = BASIC;
		}


		if(settings.console_out != NULL){ statistics.printStatsProgress((last_solver_type==DMRP?'D':'B'),numberLiterals(),conflict_count.getConflicts());}

		if(status.isKnown()) break;
	}


	if(status.isUnknown()){
		asrt(timeout >= 0 && timeout <= statistics.cpuUserTime());
		if(settings.console_out != NULL)
			settings.console_out->writeString("|!!! Timeout of %i seconds exceeded !!! ",(int)timeout);
			return UNKNOWN_TrBool;
	}

	asrt(!status.isTrue() || debugVerifyModel());

	if(settings.console_out != NULL){
		settings.console_out->writeString("|____________________________________________________________________________________________");
		settings.console_out->writeString("| ");
	}

	return status;
}



void 	DMRPSolver::clearAssignmentsKeepHooks(const bool care_for_responses){


	//asrt(care_for_responses || decisionLevel() != 0);

	if(decisionLevel() != 0){
		vector<std::pair<Literal, Clause* > > link_later;
		queue_begin 	= decision_level_end[0];

	    for(int32_t c 	= assign_stack.size()-1; c >= (int32_t)queue_begin; --c){
	    	Literal lit	= assign_stack[c];
	    	asrt(!litIsUnknown(lit));
	        Variable x  = lit.toVar();

	        if(care_for_responses){
	        	bool lit_was_hooked = litIsHooked(lit);
	        	varClearAssignmentSetPrevAndHook(x);
	        	if(!lit_was_hooked){
	        		vector<Clause*>& clses = litHookResponse(lit);
	        		warrant(findNewHookResponses(clses,link_later));
	        	}
	        }else{
	        	varClearAssignmentSetPrevAndHook(x);
	        	hook_responses[c].clear();
	        	dec_var_count[c] = 0;
	        }

	        reinsertVarOrder(x);
	    }

	    assign_stack            .resize(queue_begin);
	    decision_level_end    	.clear();

		for(uint i = 0; i != link_later.size();++i){
			asrt(litIsHooked(link_later[i].first));
			litHookResponse(link_later[i].first).push_back(link_later[i].second);
		}


		if(++clause_stamp > Clause::MAX_STAMP){
	    	clsClearAllStamps();
	    	clause_stamp = 0;
	    	++statistics.nbr_expensive_cls_stamp;
	    }
	}
}

TrBool  DMRPSolver::dmrpSolveConsiderTimeAndConflicts(uint32_t timeout, uint64_t rel_max_confl){

	asrt(debugTestNumberUnwatched(unwatched_clauses.size()));
	asrt(debugNumberHookResponsiblesAllRight());

	TrBool status 	= UNKNOWN_TrBool;

	while(statistics.cpuUserTime() < timeout){

		Clause* sat_next_cls = NULL;
		do{
			if(unwatched_clauses.empty()){			return TRUE_TrBool;}

			sat_next_cls 	= 	unwatched_clauses	.back();
			unwatched_clauses						.pop_back();
			asrt(sat_next_cls != NULL);
			if(sat_next_cls->isLearnt())			--nbr_learnt_uw_clses;

			sanity(debugOneLitIsHookedForWatchedClauses(fix_cls));

			if(!clsReLink(sat_next_cls,true)){		return TrBool(state_ok = false);}

		}while(clsLinkHook(*sat_next_cls));


		asrt(sat_next_cls != NULL && checkSatByHook(*sat_next_cls) == INVALID_LIT);

		status  = dmrpSearchSatClause(sat_next_cls,conflict_count.getConflicts() + rel_max_confl);


	    //if(settings.console_out != NULL){ statistics.printStatsProgress();}
		//print_out("Problem Clauses: %d, max_conflicts %d, delta_max_confl %d \n",unwatched_clauses.size()+1,max_conflicts, search_max_confl);

	    if(status.isFalse())	return TrBool(state_ok = false);

	    if(status.isUnknown()){
	    	//backjumpKeepLevel(0,BACKJUMP_SET_PREV);
	    	clsUnlinkWatched(sat_next_cls,true);

	    	//costly__asrt(debugOneLitIsHookedForWatchedClauses(fix_cls,sat_next_cls)); geht nur wenn vorher backjump ->0
	    	sanity(debugNumberHookResponsiblesAllRight());
	    	return UNKNOWN_TrBool;
	    }

	    asrt(status.isTrue());

	    //print_out("reduced one to %d\n",unwatched_clauses.size());
    	clearAssignmentsKeepHooks(true);
    	warrant(clsLinkHook(*sat_next_cls));

    	sanity(debugOneLitIsHookedForWatchedClauses(fix_cls));
    	sanity(debugNumberHookResponsiblesAllRight());
    }

	return UNKNOWN_TrBool;
}


TrBool DMRPSolver::dmrpSearchSatClause(Clause* cls, const uint32_t abs_max_confl){


	asrt(state_ok && decisionLevel() == 0);
	asrt(delta_stack.empty() );
	asrt(abs_max_confl >= conflict_count.getConflicts());

	uint64_t max_learnts	= conflict_count.getMaxLearnts();
	TrBool   result			= UNKNOWN_TrBool;

	asrt(debugNumberHookResponsiblesAllRight());
	//asrt(sum_element_sizes(var_makes) == 0);

	deltaProbClausesEnsureLevel(0);
	addProbClsDelta(*cls,0);

	// Reduce the set of learnt clauses:
	if(numberLearntLongClses() >= max_learnts +numberAssignments()){
		dmrpLearntsGC(cls,((max_learnts +numberAssignments())/2)+numberLiteralsInShortClses());
		//print_out("Number of learnts now %d \n",numberLearntLongClses());
	}

	asrt(debugAllWatchedNonDeltaClausesHooked());

	while(conflict_count.getConflicts() < abs_max_confl || decisionLevel() == 0 ){

		sanity(debugNumberHookResponsiblesAllRight());
    	uint32_t q_begin		= queue_begin;
    	int32_t bcp_res    		= bcp(INIT);

    	if(bcp_res == BCP_SUCCESS){         //there is no conflict

    		if(decisionLevel() != 0){

    		    asrt(delta_stack.size() <= decisionLevel() + 1);
    		    deltaProbClausesEnsureLevel(decisionLevel());
    			updateDelta(q_begin);

    			//print_out("Props per decision: %7.2f \r",(double)numberOfAssignments()/(double)decisionLevel());
    		}else {
    			unlinkUnhookedClauses(q_begin);
   				sanity(debugNumberHookResponsiblesAllRight());

    			asrt(debugAllWatchedNonDeltaClausesHooked());

    			if(q_begin + 1< queue_begin){ //only check if propagation causes at least one other literal
    				if(!checkUnwatchedClauses()){ result =  TrBool(state_ok = false); break;}
    			}


    			asrt(delta_stack[0]->size() == 1);
    			asrt(delta_stack[0]->deltasBegin()->getClause() == cls);

    			asrt(debugOneLitIsHookedForWatchedClauses(fix_cls,cls));


    		    asrt(debugAtLeastTwoUnassignedLits(*cls,0));
    		    if(checkSatByHook(*cls) != INVALID_LIT){
    		    	//print_out("Klausel durch neuen Hook erfüllt \n");
    		    	result =  TRUE_TrBool;  break;
    		    }
    		}

    		sanity(debugAllWatchedNonDeltaClausesHooked());

			// New decision:
			Literal next    = nextDeltaDecision();

			/*if(decisionLevel() == 0){
				print_out("Decision at 0 is %d with break count %u \n",next.toDIMACS(),litHookResponse(next.flipped()).size());
				delta_stack[0][0]->getClause()->print(*settings.console_out);
			}*/

			//check if model is found
			if (next == INVALID_LIT){	result =  TRUE_TrBool;  break; }

			incDecCount(next);


            // Increase decision level and enqueue 'next'
            asrt(litIsUnknown(next));
            newDecisionLevel();
            uncheckedEnqueue<Clause*>(next);

            set_max_int(statistics.max_dec_level,decisionLevel());


           // print_out("Entscheidung auf level %d Literal %d \r",decisionLevel(),next.toDIMACS());



    	}else{                         //there is a conflict

    		 conflict_count.incConflict();

    		 asrt(bcp_res != 1);
    		 if (!firstUIP(bcp_res))  {                          result = TrBool(state_ok = false); 	break;}
    		 asrt(litAssignLevel(conflict_lemma[0]) == decisionLevel());

             reduceSelfImpl();
             asrt(debugAtLeastOneHooked(conflict_lemma,0));

             int32_t backtrack_level = backtrackLevel();

             statistics.storeBackjumpingDistance(decisionLevel()- backtrack_level);


             backjumpKeepLevel(backtrack_level,BACKJUMP_SET_PREV);
             asrt(litIsUnknown(conflict_lemma[0]));
             restoreProbClsDelta(backtrack_level);

             if(backtrack_level == 0){
            	 max_learnts	= conflict_count.getMaxLearnts();
            	 if(numberLearntLongClses() >= max_learnts +numberAssignments()){
            	 	dmrpLearntsGC(cls,((max_learnts +numberAssignments())/2)+numberLiteralsInShortClses());
            	 	//print_out("Number of learnts now %d \n",numberLearntLongClses());
            	 }

            	 //panic too many literals

				 if ( numberLiterals() > settings.max_literals){
					//print_out("Force DMRP Learnts GC due to Memory: %llu Literals, and %llu allowed\n",numberLiterals(),settings.max_literals);
					dmrpLearntsGC(cls,numberLearntLongClses() / 3);
					break;
				 }
             }

             asrt(litIsUnknown(conflict_lemma[0]));

             const uint32_t l_sz(conflict_lemma.size());
             if(l_sz > 2){
            	 asrt(l_sz <= Clause::MAX_SIZE);
            	 Clause* learnt_cl = clsCreate(conflict_lemma.begin(),conflict_lemma.end(),true);
            	 asrt(debugAtLeastOneHooked(*learnt_cl));
            	 clsBump(learnt_cl);
            	 clsLinkHookNewLearnt(*learnt_cl);
            	 uncheckedEnqueue(conflict_lemma[0], learnt_cl);
             }else if(l_sz == 2){
             	clsCreateBinary(conflict_lemma[0],conflict_lemma[1]);
                uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1]);
             }else{
             	uncheckedEnqueue<Clause*>(conflict_lemma[0]);
             }


             var_act_hp.decayAll();
             claDecayActivity();
        }
    }

    clearDeltaClausesAndMakeCount();
	//print_out("return result %s \n",(result.isKnown()?(result.isTrue()?"true":"false"):"unknown"));
	return result;
}

bool	DMRPSolver::computeHooksEnsureBinariesSAT(){

	TrBool 	status = search( UNLINK,max64);
	asrt(status.isKnown());

	if(status.isFalse()) return false;

	//in case not all variables exist -> clear the hook responses
	for(uint i = assign_stack.size(); i < hook_responses.size();++i){dec_var_count[i] = 0; hook_responses[i].clear();}

	clearAssignmentsKeepHooks(false);

	for(uint i = queue_begin;i--;){dec_var_count[i] = 0; varSetHook(assign_stack[i]);hook_responses[i].clear();}

	asrt(debugAllHookResponsesCleared());
	asrt(debugAllDecCountsZero());

	initHookResponses(fix_cls);
	initHookResponses(learnt_cls);

	asrt(debugTestNumberUnwatched(unwatched_clauses.size()));
	asrt(debugNumberHookResponsiblesAllRight());

	if(!checkUnwatchedClauses()) return false;

	unwatched_clauses.sort(ClActAscendingFunctor());

	return true;

}

uint32_t DMRPSolver::computeMakeCount(Literal lit, uint32_t lower_bound, Clause* except){

	asrt(litIsUnknown(lit) && !litIsHooked(lit));
	vector<MakeClause>& make_cands = var_makes[lit.toVar().idx()];
	uint32_t new_sz = make_cands.size();
	uint32_t mcount = 0;

	uint32_t idx = 0;
	while(idx < new_sz && (mcount + new_sz - idx) >= lower_bound){
		MakeClause& mcls = make_cands[idx];

		if(!mcls.level_clauses->isValid()){
			make_cands[idx] = make_cands[--new_sz];
			continue;
		}

		DeltaClause& dcls = mcls.level_clauses->getDeltaAt(mcls.index);
		asrt(dcls.getClause() != NULL);

		if(dcls.getClause() != except && !litIsTrue(dcls.getLiteral())){

			if(lit = checkSatByLit(*dcls.getClause()), lit == INVALID_LIT){
				++mcount;
			}else{
				dcls.setLiteral(lit);
			}
		}
		++idx;
	}

	if(new_sz < make_cands.size()) make_cands.resize(new_sz);
	return mcount;
}


bool   DMRPSolver::collectPossibleDecLits(vector<Literal>& possible_lits,DeltaClause& dcls){

	asrt(dcls.getClause() != NULL);
	Clause& 	cls		= *dcls.getClause();
	asrt(litIsUnknown(cls[0]) || litIsUnknown(cls[1]) || litIsTrue(cls[0]) || litIsTrue(cls[1]) );

	possible_lits.clear();

	for(uint32_t lidx=0;lidx != cls.size();++lidx){
		Literal cand_lit = cls[lidx];
		if(litIsUnknown(cand_lit)){
			asrt(!litIsHooked(cand_lit));
			possible_lits.push_back(cand_lit);
		}else if(litIsTrue(cand_lit)){
			dcls.setLiteral(cand_lit);
			return true;
		}
	}
	return false;
}



Literal DMRPSolver::nextDeltaDecision(){
	//todo bessere entscheidung auf niedrigeren leveln

	sanity(debugAllWatchedNonDeltaClausesHooked());

	vector<Literal> possible_lits;
	//var_act_t best_act(0);

	asrt(delta_stack_idx <= delta_stack.size());


	while(delta_stack_idx--){
		OneLevelDeltaClauses* ldcls = delta_stack[delta_stack_idx];
		asrt(ldcls->isValid());

		for(delta_riter lx = ldcls->deltasRBegin(); lx != ldcls->deltasREnd();++lx){
			DeltaClause& dcls = *lx;

			if(litIsTrue(dcls.getLiteral()) || collectPossibleDecLits(possible_lits,dcls)) continue;

			uint32_t plstart = 0;
			asrt(possible_lits.size() > 1);

			uint32_t 	mcount	= 0;

			#ifdef ON_TIE_DECISION_COUNT
			uint16_t	dcount	= MAX16;
			#elif defined ON_TIE_ACTIVITY
			var_act_t   max_act = var_act_t(0);
			#elif defined ON_TIE_BREAK_COUNT
			uint32_t	bcount	= numberAllLongClses()+1;
			#endif

			if(settings.randy.nextDouble() > 0.02){

				sort(possible_lits.begin(),possible_lits.end(),MakeCountDescendingFunctor(this));

				plstart		= possible_lits.size();
				Clause* cls = dcls.getClause();

				for(uint i = 0; i < plstart;++i){
					Literal lit = possible_lits[i];
					if(var_makes[lit.toVar().idx()].size() < mcount) break; //since vec is sorted

					uint32_t cand = computeMakeCount(lit,mcount,cls);
					if(cand > mcount){
						mcount 		= cand;

						#ifdef ON_TIE_DECISION_COUNT
						dcount		= getDecCount(lit);
						#elif defined ON_TIE_ACTIVITY
						max_act		= var_act_hp.getValue(lit.toVar());
						#elif defined ON_TIE_BREAK_COUNT
						bcount		= litHookResponse(lit.flipped()).size();
						#endif

						possible_lits.resize(plstart);
						possible_lits.push_back(lit);
					}else if(cand == mcount){

						#ifdef ON_TIE_DECISION_COUNT
						uint16_t dc = getDecCount(lit);
						if(dc < dcount){
							dcount	= dc;
							possible_lits.resize(plstart);
							possible_lits.push_back(lit);
						}else if(dc == dcount){
							possible_lits.push_back(lit);
						}

						#elif defined ON_TIE_ACTIVITY
						var_act_t a = var_act_hp.getValue(lit.toVar());
						if(a > max_act){
							max_act = a;
							possible_lits.resize(plstart);
							possible_lits.push_back(lit);
						}else if(a == max_act){
							possible_lits.push_back(lit);
						}

						#elif defined ON_TIE_BREAK_COUNT
						uint16_t bc = litHookResponse(lit.flipped()).size();
						if(bc < bcount){
							bcount	= bc;
							possible_lits.resize(plstart);
							possible_lits.push_back(lit);
						}else if(bc == bcount){
							possible_lits.push_back(lit);
						}
						#endif
					}
				}

			}

			const uint interval = possible_lits.size() - plstart;
			asrt(interval > 0);

			Literal ret_lit = interval == 1 ? possible_lits[plstart]: possible_lits[plstart + settings.randy.nextInt(interval)];
			asrt(ret_lit != INVALID_LIT);
			dcls.setLiteral(ret_lit);
			++delta_stack_idx;
			return ret_lit;
		}
	}

	return INVALID_LIT;
}


bool    DMRPSolver::findNewHookResponses(vector<Clause*>& clses, vector<std::pair<Literal,Clause*> >& link_later){


	bool found_for_all = true;

	for(cls_iter_t iter = clses.begin(); iter != clses.end();++iter){
		Clause* cls = *iter;
		bool  found_response = false;

		for(uint c = 0; c != cls->size(); ++c){
			const Literal lit = (*cls)[c];
			if(litIsNotFalse(lit) && litIsHooked(lit)){
				litHookResponse(lit).push_back(cls);
				found_response = true; break;
			}
			if(litIsTrue(lit)){
				asrt(!litIsHooked(lit));
				link_later.push_back(std::pair<Literal,Clause*>(lit,cls));
				found_response = true; break;
			}
		}

		if(!found_response){
			//print_out("extend unwatched clauses\n");
			clsUnlinkWatched(cls,true);
			clsBump(cls);
			found_for_all = false;
		}
	}

	clses.clear();

	return found_for_all;
}

void	DMRPSolver::clearDeltaClausesAndMakeCount(){
	uw_clses_for_mc.invalidate();

	for(uint i = all_make_vars.size();i--;){
		Variable var = all_make_vars[i];
		var_makes[var.idx()].clear();
	}
	all_make_vars.clear();

	clearDeltaStack(0,true);
	delta_stack.clear();
	delta_stack_idx = 0;

	for(uint i = rubbish_level_delta_clauses.size();i--;){delete rubbish_level_delta_clauses[i];}
	rubbish_level_delta_clauses.clear();
	uw_clses_for_mc.reValidate();
}

/*
uint	DMRPSolver::removeRubbishDeltaClauses(vector<DeltaClause*>& dclses,const bool del){
	vector<DeltaClause*>::iterator from,to,end;
	end = dclses.end();
	for(from = to = dclses.begin();from != end;++from){
		if(!(*from)->isRubbish()){*to++ = *from;}
		else if(del) {delete *from;}
	}
	if(to != end){dclses.erase(to,end);}
	return end - to;
}*/

bool	DMRPSolver::removeRubbishMakeCounts(Variable var){

	vector<MakeClause>& make_cands = var_makes[var.idx()];
	uint32_t new_sz = make_cands.size();
	uint32_t idx = 0;

	while(idx < new_sz){
		MakeClause& mcls = make_cands[idx];

		if(mcls.level_clauses->isValid()){ ++idx;}
		else{ make_cands[idx] = make_cands[--new_sz];}
	}

	if(new_sz < make_cands.size()) make_cands.resize(new_sz);
	return new_sz != 0;
}


void 	DMRPSolver::removeRubbishDeltaClauses(const bool clear_global){

	if(settings.global_make_count && clear_global) uw_clses_for_mc.invalidate();

	vector<Variable>::iterator from,to;
	for(from = to = all_make_vars.begin(); from !=  all_make_vars.end();++from){
		if(removeRubbishMakeCounts(*from)){*to++ = *from;}
	}
	if(to != all_make_vars.end()) all_make_vars.erase(to,all_make_vars.end());


	for(uint32_t i = rubbish_level_delta_clauses.size();i--;){
		asrt(!rubbish_level_delta_clauses[i]->isValid());
		delete rubbish_level_delta_clauses[i];
	}
	rubbish_level_delta_clauses.clear();

	if(settings.global_make_count && clear_global) uw_clses_for_mc.reValidate();
}



void 	DMRPSolver::updateDeltaCheckClause(cls_iter_t& from,cls_iter_t& copy){
	Clause& cl = **from++;

	Literal lit = cl[0];
	if(!litIsFalse(lit) && litIsHooked(lit)){litHookResponse(lit).push_back(&cl);return ;}
	if(litIsTrue(lit)){	*copy++ = &cl;return ;}

	lit = cl[1];
	if(!litIsFalse(lit) && litIsHooked(lit)){litHookResponse(lit).push_back(&cl);return ;}
	if(litIsTrue(lit)){	*copy++ = &cl;return ;}


	for(uint k = cl.getNextK(clause_stamp); k != cl.size(); ++k){
		lit = cl[k];
		if(!litIsFalse(lit) && litIsHooked(lit)){litHookResponse(lit).push_back(&cl);return ;}
		if(litIsTrue(lit)){	*copy++ = &cl;return ;}
	}
	*copy++ = &cl;
	addProbClsDelta(cl,decisionLevel());
}


void 	DMRPSolver::updateDelta(uint32_t q_begin){

	sanity(debugNumberHookResponsiblesAllRight());


	cls_iter_t iter,copy,end;

	while(q_begin < assign_stack.size()){
		const Literal true_lit             	= assign_stack[q_begin++];

		if(litIsHooked(true_lit)) continue;
	    vector<Clause*>& responses    		= litHookResponse(true_lit.flipped());

	    iter = copy =  responses.begin();
	    end     	= responses.end();

	    while(iter != end){
	    	updateDeltaCheckClause(iter,copy);
	    }
	    if(copy != end){responses.erase(copy,end);}
	 }
	sanity(debugAllWatchedNonDeltaClausesHooked());
}



void 	DMRPSolver::unlinkUnhookedClauses(uint32_t q_begin){


	asrt(decisionLevel() == 0);
	asrt(queue_begin == assign_stack.size());
	asrt(debugNumberHookResponsiblesAllRight());

	vector<std::pair<Literal,Clause*> > link_later;
	while(q_begin < queue_begin){
		const Literal true_lit			= assign_stack[q_begin++];

		if(litIsHooked(true_lit)) { 	continue;}

		vector<Clause*>& responses    	= litHookResponse(true_lit.flipped());

		//flip hook and find new responses
		varSetHook(true_lit);

		findNewHookResponses(responses,link_later);
	}

	for(uint i = 0; i != link_later.size();++i){
		asrt(litIsHooked(link_later[i].first));
		litHookResponse(link_later[i].first).push_back(link_later[i].second);
	}

	//print_out("Sum hooks resp: %d, sum linked cls: %d\n",sum_element_sizes(hooks_responses),(numberOfLongOrigClauses() + numberOfLongLearntClauses() - unwatched_clauses.size()));
	asrt(debugNumberHookResponsiblesAllRight());
}





void  DMRPSolver::clsLinkHookNewLearnt(Clause& cl){


	asrt(litIsUnknown(cl[0]));
	if(litIsHooked(cl[0])){
		litHookResponse(cl[0]).push_back(&cl);
		return;
	}

	//search deepest hooked lit that is assigned false
	uint32_t level = 0;
	Literal take_response = INVALID_LIT;
	for(uint32_t lidx = 1; lidx != cl.size(); ++lidx) {
		asrt(litIsFalse(cl[lidx]));
		if(litIsHooked(cl[lidx]) && litAssignLevel(cl[lidx]) > level){
			take_response = cl[lidx];
			level = litAssignLevel(take_response);
		}
	}

	asrt(level != 0);
	asrt(take_response != INVALID_LIT);
	litHookResponse(take_response).push_back(&cl);
	addProbClsDelta(cl,level);
}



void	DMRPSolver::addProbClsDelta(Clause& cl,const uint level){

	asrt(debugAtLeastTwoUnassignedLits(cl,level));
	asrt(delta_stack.size() == decisionLevel()+ 1);
	asrt(delta_stack[0]->size()<2);
	OneLevelDeltaClauses* ldcls = delta_stack[level];
	asrt(ldcls != NULL && ldcls->isValid());

	const uint32_t deltaClauseIndex = ldcls->addClauseGetIndex(&cl);

	for(uint i = cl.size();i--;){
		asrt(!litIsTrue(cl[i]));
		if(litIsUnknown(cl[i]) || litAssignLevel(cl[i]) > level){
			asrt(!litIsHooked(cl[i]));
			incrementMakeCount(cl[i],ldcls,deltaClauseIndex);}
	}

}


void	DMRPSolver::clearDeltaStack(uint32_t start_idx, const bool del){

	if(del){
		while(start_idx < delta_stack.size()){delete delta_stack[start_idx++];}
	}else{
		while(start_idx < delta_stack.size()){
			OneLevelDeltaClauses* dclses	= delta_stack[start_idx++];
			dclses							->invalidate();
			rubbish_level_delta_clauses		.push_back(dclses);
		}
	}


}


void	DMRPSolver::restoreProbClsDelta(uint32_t level){

	asrt(delta_stack.size() >= level + 1);

	clearDeltaStack(level + 1,false);

	deltaProbClausesEnsureLevel(level);

	if(rubbish_level_delta_clauses.size() > 250000){
		removeRubbishDeltaClauses(false);
		//print_out("Removed rubbish delta clauses!\n");
	}


	sanity(debugAllWatchedNonDeltaClausesHooked());
}


void DMRPSolver::initHookResponses(vector<Clause*>& clauses){

	for(cls_iter_t iter = clauses.begin(); iter != clauses.end();++iter){
		Clause& cl = **iter;
		asrt(!cl.isRubbish());

		if(cl.isWatched()){
			asrt(	(litIsHooked(cl[0]) && litIsNotFalse(cl[0])) ||
						(litIsHooked(cl[1]) && litIsNotFalse(cl[1]) ));
			Literal lit = (litIsHooked(cl[0]) && litIsNotFalse(cl[0])) ? cl[0]: cl[1];
			litHookResponse(lit).push_back(&cl);
		}
	}
}

void DMRPSolver::dmrpLearntsGC(Clause* excep,uint32_t new_sz){


	asrt(decisionLevel() == 0);

	if(new_sz >=  numberLearntLongClses()) return;

	cls_act_t act 	= excep->getActivity();
	excep->getActivity() = settings.cls_max_act;

	learntsGC(new_sz,false);
	excep->getActivity() = act;
	asrt(!excep->isRubbish() && excep->isWatched());

	cls_iter_t iter,to,end;
	for(uint32_t i  = 0; i != hook_responses.size() ; ++i){
		iter = to 	= hook_responses[i].begin();
		end			= hook_responses[i].end();
		while(iter != end){
			Clause* cls = *iter++;
			if(!cls->isRubbish()){*to++ = cls;}
		}
		if(to != end){hook_responses[i].erase(to,end);}
	}
}





bool DMRPSolver::clsLinkHook(Clause& cl){

	for(uint32_t lidx = 0; lidx != cl.size(); ++lidx) {
		asrt(varHook(cl[lidx].toVar()) != UNKNOWN_TrBool);
		if(litIsHooked(cl[lidx]) && litIsNotFalse(cl[lidx])){
			litHookResponse(cl[lidx]).push_back(&cl);
			return true;
		}
	}

	//print_out("Keine Hooks f�r die%s Klausel: \n",(cl.isLearnt()?" gelernte":""));
	//cl.print(*settings.console_out);

	return false;
}


bool	DMRPSolver::bcpDmrpOK(){

	asrt(decisionLevel() 	== 0);
	asrt(queue_begin 		<= assign_stack.size());
	uint32_t q_begin 		= queue_begin;

	if(bcp(RETURN) != BCP_SUCCESS){ return false;}

	unlinkUnhookedClauses(q_begin);
	sanity(debugNumberHookResponsiblesAllRight());

	return  true;
}



bool	DMRPSolver::checkUnwatchedClauses(){


	sanity(debugNumberHookResponsiblesAllRight());

	asrt(decisionLevel() == 0);

	removeRubbishDeltaClauses(true);

	uint32_t q_begin 		= queue_begin;
	const bool add_global 	= settings.global_make_count;

	do{
		sanity(debugNumberHookResponsiblesAllRight());

		state_ok = bcpDmrpOK();
		if(!state_ok){ return false;}

		q_begin = queue_begin;
		uwc_iter 	iter 	= unwatched_clauses.begin();

		while(iter != unwatched_clauses.end()){
			Clause* cls = *iter;
			if(checkSatByHook(*cls) != INVALID_LIT){
				if(!clsReLink(cls,true)){ return state_ok = false;}
				warrant(clsLinkHook(*cls));
				iter = unwatched_clauses.erase(iter);
				if(cls->isLearnt())	--nbr_learnt_uw_clses;
			}else{
				uint8_t valid_lits = clsValidLitsToFront(*cls);
				if(valid_lits == 0) return (state_ok = false);

				if(valid_lits == 1){
					asrt(!litIsFalse((*cls)[0]));

					if(litIsUnknown((*cls)[0])){
						sanity(debugNumberHookResponsiblesAllRight());

						uncheckedEnqueue<Clause*>((*cls)[0]);
						state_ok = bcpDmrpOK();
						if(!state_ok) return false;
					}
					asrt(litIsTrue((*cls)[0]));
				}

				//add to make counts
				if(add_global) {
					const uint32_t deltaClauseIndex = uw_clses_for_mc.addClauseGetIndex(cls);
					for(uint i = cls->size();i--;){			asrt(!litIsTrue((*cls)[i]));
						if(litIsUnknown((*cls)[i])){		asrt(!litIsHooked((*cls)[i]));
							incrementMakeCount((*cls)[i],&uw_clses_for_mc,deltaClauseIndex);}
					}
				}

				++iter;
			}

		}

	}while(q_begin != queue_begin);

	//print_out("Sum hooks resp: %d, sum linked cls: %d\n",sum_element_sizes(hooks_responses),(numberOfLongOrigClauses() + numberOfLongLearntClauses() - unwatched_clauses.size()));
	asrt(debugNumberHookResponsiblesAllRight());

	//print_out("End check unwatched %d\n",unwatched_clauses.size());
	return true;
}


/*
void	DMRPSolver::resetNormalClsState(){
	clause_stamp = 0;
	for(cls_iter_t iter = fix_cls.begin(); iter != fix_cls.end();++iter){(*iter)->clearStampAndFlag();}
	for(cls_iter_t iter = learnt_cls.begin(); iter != learnt_cls.end();++iter){(*iter)->clearStampAndFlag();}
}
*/






bool 	DMRPSolver::relinkAllUnwatchedClauses(){

	asrt(decisionLevel() == 0);
	//var_act_hp.decayAll();
	for(uwc_iter iter = unwatched_clauses.begin();iter!= unwatched_clauses.end();++iter){
		Clause* cls = *iter;
		if(!clsReLink(cls,false)) return state_ok = false;
		/*if(checkSatByHook(*cls) == INVALID_LIT){
			//print_out("Heavy Broken %s Clause: ",(cls->isLearnt()?"learnt":"original"));
			for(uint32_t i = cls->size(); i--;) {
				//print_out(" %d ",(*cls)[i].toDIMACS());
				varBump((*cls)[i].toVar());

				//debugPrintHookResponses((*cls)[i].flipped());

			}
			cls->setFilterFlag();
			//print_out("  \n");
		}*/
	}
	unwatched_clauses.clear();
	nbr_learnt_uw_clses = 0;
	return true;
}

bool	DMRPSolver::clsPlaceValidLitToFront(Clause& cl, const uint32_t front,uint32_t& start){

	for(;start < cl.size();++start ){
		if(!litIsFalse(cl[start])){
			if(front != start){Literal t = cl[front];cl[front]=cl[start];cl[start]=t;}
			++start;
			return true;
		}
	}
	return false;
}

uint8_t DMRPSolver::clsValidLitsToFront(Clause& cl){

	asrt(decisionLevel() == 0);
	uint32_t k = 0;
	if(!clsPlaceValidLitToFront(cl,0,k))	return  0;

	if(!clsPlaceValidLitToFront(cl,1,k))	return 1;

	return 2;
}

bool    DMRPSolver::clsReLink(Clause* cls,const bool care_for_hooks){

	asrt(decisionLevel() == 0);

	//sort(cls->litsBegin(),cls->litsEnd(),LitActDescendingFunctor(this));

	uint8_t valid_lits = clsValidLitsToFront(*cls);
	if(valid_lits == 0) return state_ok = false;

	if(valid_lits == 1){
		asrt(!litIsFalse((*cls)[0]));

		if(litIsUnknown((*cls)[0])){
			asrt(state_ok);
			uncheckedEnqueue<Clause*>((*cls)[0]);

			if(care_for_hooks){
				sanity(debugNumberHookResponsiblesAllRight());
				state_ok = bcpDmrpOK();
			}else{
				state_ok = bcp(RETURN) == BCP_SUCCESS;
			}


			if(!state_ok){ return state_ok = false;}
		}
	}

	clsLinkWatched(cls);
	return true;
}




const char*  dmrp_solver_cpp_version(){return "DMRPSolver.cpp:     $Revision: 1.7 $";}

